﻿//Copyright (c) Microsoft Corporation.  All rights reserved.

using Microsoft.WindowsAzure.Storage;
using Microsoft.WindowsAzure.Storage.Table;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Reflection;
using System.Threading;

namespace Microsoft.Ted.Wacel.TableStorage
{
    public class CloudTableStorage<E> where E : new()
    {
        private CloudTableClient mClient;
        private CloudTable mTable;

        private PropertyInfo mIdField;
        private PropertyInfo mPartitionField;

        private bool mThrowExceptionWhenNotFound;

        public CloudTableStorage(string idFieldName, string connectionString, string tableName, bool notfoundException = true)
            :this(idFieldName,idFieldName, connectionString, tableName, notfoundException)
        {
        }
        public CloudTableStorage(string partitionFieldName, string idFieldName, string connectionString, string tableName, bool notfoundException = true)
        {
            mThrowExceptionWhenNotFound = notfoundException;

            Type type = typeof(E);
            mIdField = type.GetProperties().FirstOrDefault(f => f.Name == idFieldName);
            if (mIdField == null)
                throw new ArgumentException("Attribute not found on object.", idFieldName);

            mPartitionField = type.GetProperties().FirstOrDefault(f => f.Name == partitionFieldName);
            if (mPartitionField == null)
                throw new ArgumentException("Attribute not found on object.", partitionFieldName);

            CloudStorageAccount account = CloudStorageAccount.Parse(connectionString);
            mClient = account.CreateCloudTableClient();
            mTable = mClient.GetTableReference(tableName);
            try
            {
                mTable.CreateIfNotExists();
            }
            catch (StorageException exp)
            {
                if (exp.RequestInformation.HttpStatusCode != (int)HttpStatusCode.Conflict)
                    throw;
            }
        }

        public void Clear()
        {
            mTable.DeleteIfExists();
            while (true)
            {
                try
                {
                    mTable.CreateIfNotExists();
                    return;
                }
                catch (StorageException exp)
                {
                    if (exp.RequestInformation.HttpStatusCode != (int)HttpStatusCode.Conflict)
                        throw;
                    else
                        Thread.Sleep(1000);
                }
            }
        }

        private AutoDynamicTableEntity<E> makeEntity(E entity)
        {
            var partitionProperty = mPartitionField.GetValue(entity);
            string partition = partitionProperty == null ? null : partitionProperty.ToString();
            var rowProperty = mIdField.GetValue(entity);
            string row = rowProperty == null ? null : rowProperty.ToString();
            if (string.IsNullOrEmpty(row) && string.IsNullOrEmpty(partition))
            {
                row = "K" + Guid.NewGuid().ToString("N");
                partition = row;
            }
            else if (string.IsNullOrEmpty(row))
                row = "K" + Guid.NewGuid().ToString("N");
            else if (string.IsNullOrEmpty(partition))
                partition = row;
            return new AutoDynamicTableEntity<E>(entity, partition, row);
        }
        public void Insert(E entity)
        {
            TableOperation operation = TableOperation.Insert(makeEntity(entity));
            mTable.Execute(operation);
        }
        public void InsertOrReplace(E entity)
        {
            TableOperation operation = TableOperation.InsertOrReplace(makeEntity(entity));
            mTable.Execute(operation);
        }
        public void BulkInsertOrReplace(IEnumerable<E> entities)
        {
            TableBatchOperation batch = new TableBatchOperation();
            if (entities.Count() > 0)
            {
                foreach (var entity in entities)
                    batch.InsertOrReplace(makeEntity(entity));
                mTable.ExecuteBatch(batch);
            }
        }
        public void InsertOrMerge(E entity)
        {
            TableOperation operation = TableOperation.InsertOrMerge(makeEntity(entity));
            mTable.Execute(operation);
        }
        public void Replace(E entity)
        {
            var newEntity = makeEntity(entity);
            newEntity.ETag = "*";
            TableOperation operation = TableOperation.Replace(newEntity);
            mTable.Execute(operation);
        }
        public void Update(E entity)
        {
            string Etag = "";
            var oldEntity = Read(mIdField.GetValue(entity).ToString(), out Etag, partitionId:mPartitionField.GetValue(entity).ToString());
            var newEntity = makeEntity(entity);
            newEntity.ETag = Etag;
            TableOperation operation = TableOperation.InsertOrMerge(newEntity);
            mTable.Execute(operation);
        }
        public void Delete(string id, string tenantId = null, string partitionId = null)
        {
            if (string.IsNullOrEmpty(id))
                throw new ArgumentNullException("Id can't be null.");
            if (!string.IsNullOrEmpty(tenantId))
                throw new ArgumentException("Tenant Id is not supported.");
            if (string.IsNullOrEmpty(partitionId))
                partitionId = id;

            TableOperation readOperation = TableOperation.Retrieve<AutoDynamicTableEntity<E>>(partitionId, id);
            TableResult readResult = mTable.Execute(readOperation);
            if (readResult.Result != null)
            {
                TableOperation operation = TableOperation.Delete((AutoDynamicTableEntity<E>)readResult.Result);
                mTable.Execute(operation);
            }
        }
        public E Read(string id, string tenantId = null, string partitionId = null)
        {
            if (string.IsNullOrEmpty(id))
                throw new ArgumentNullException("Id can't be null.");
            if (!string.IsNullOrEmpty(tenantId))
                throw new ArgumentException("Tenant Id is not supported.");
            if (string.IsNullOrEmpty(partitionId))
                partitionId = id;

            TableOperation readOperation = TableOperation.Retrieve<AutoDynamicTableEntity<E>>(partitionId, id);
            TableResult readResult = mTable.Execute(readOperation);
            if (readResult.Result != null)
                return ((AutoDynamicTableEntity<E>)readResult.Result).POCO;
            else
            {
                if (mThrowExceptionWhenNotFound)
                    throw new ArgumentOutOfRangeException(string.Format("Entity with parition:'{0}', row:'{1}' not found.", partitionId, id));
                else
                    return default(E);
            }
        }
        public E Read(string id, out string Etag, string tenantId = null, string partitionId = null)
        {
            if (string.IsNullOrEmpty(partitionId) || string.IsNullOrEmpty(id))
                throw new ArgumentNullException("Id can't be null.");
            if (!string.IsNullOrEmpty(tenantId))
                throw new ArgumentException("Tenant Id is not supported.");

            TableOperation readOperation = TableOperation.Retrieve<AutoDynamicTableEntity<E>>(partitionId, id);
            TableResult readResult = mTable.Execute(readOperation);
            if (readResult.Result != null)
            {
                Etag = readResult.Etag;
                return ((AutoDynamicTableEntity<E>)readResult.Result).POCO;
            }
            else
            {
                if (mThrowExceptionWhenNotFound)
                    throw new ArgumentOutOfRangeException(string.Format("Entity with parition:'{0}', row:'{1}' not found.", partitionId, id));
                else
                {
                    Etag = "";
                    return default(E);
                }
            }
        }
        public IEnumerable<E> List(Func<E, bool> predicate, string tenantId = null, string partitionId = null)
        {
            if (!string.IsNullOrEmpty(tenantId))
                throw new ArgumentException("Tenant Id is not supported.");
            TableQuery<AutoDynamicTableEntity<E>> query;
            if (string.IsNullOrEmpty(partitionId))
                query = new TableQuery<AutoDynamicTableEntity<E>>();
            else
                query = new TableQuery<AutoDynamicTableEntity<E>>().Where(TableQuery.GenerateFilterCondition("PartitionKey", QueryComparisons.Equal, partitionId));

            List<E> ret = new List<E>();

            foreach (AutoDynamicTableEntity<E> entity in mTable.ExecuteQuery(query))
            {
                mPartitionField.SetValue(entity.POCO, entity.PartitionKey);
                mIdField.SetValue(entity.POCO, entity.RowKey);
                if (predicate == null || predicate(entity.POCO))
                    ret.Add(entity.POCO);
            }

            return ret;
        }
        public  IEnumerable<E> List(string tenantId = null, string partitionId = null)
        {
            return this.List(null, tenantId, partitionId);
        }
    }
}
